# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains Pipeline class for ML flow focusing on model
building, training, validation and testing in DEV domain.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import logging
from typing import Any, Dict, Tuple
from datetime import *
from datetime import date as date

from enum import Enum
import shutil
import zipfile
import os
import hashlib
import pickle
import base64
import json
import pandas as pd

from snowflake.snowpark import Window as W
from snowflake.snowpark import DataFrame as SDF
import snowflake.snowpark.functions as F
from snowflake.snowpark.functions import col

from snowflake_ai.common import AppConfig, ConfigType, ConfigKey
from snowflake_ai.connect import AppConnect, SnowConnect
from snowflake_ai.connect import DataFrameFactory as DFF
from snowflake_ai.mlops import FlowContext, Pipeline, Step


# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s]  %(message)s'
)


class LibType(Enum):
    Default = "default"
    SKLearn = "scikit-learn"
    Pytorch = "pytorch"
    TensorFlow = "tensorflow"
    Databricks = "databricks"
    Snowflake = "snowflake"



class SnowRegistry:
    """
    This class provides Snowflake specific pipeline functionality
    """
    _logger = logging.getLogger(__name__)

    def __init__(
            self,
            reg_key: str,
            ctx: FlowContext,
            app_config: AppConfig = None
    ) -> None:
        self.logger = SnowRegistry._logger
        self.registry_key = reg_key
        self.context = ctx
        self.app_config = app_config
        self.app_configs: Dict = app_config.get_all_configs() if app_config \
                is not None else AppConfig.get_all_configs()
        self.data_connect: SnowConnect = None

        # load and init config
        self._init_config()



    def _init_config(self):
        gk, k = AppConfig.split_group_key(self.registry_key)
        reg_config: Dict = self.app_configs[ConfigType.MLOps.value][gk][k]
                
        if (reg_config[ConfigKey.TYPE.value] == AppConfig.T_CONN_SNFLK)\
                and (reg_config.get(ConfigKey.CONN_DATA.value) is not None):
            dconn_ref = reg_config.get(ConfigKey.CONN_DATA.value)
            ac = AppConnect.get_app_connects().get(dconn_ref)
            if ac is None:
                ac = SnowConnect(dconn_ref, self.app_config)
                AppConnect.get_app_connects()[dconn_ref] = ac

            self.data_connect = ac
            self.logger.debug(
                "SnowRegistry._init_config(): Initialize [{dconn_ref}] "\
                f"DataConnect for Registry [{self.registry_key}]."
            )
        
        # get registry table
        self.registry_table = reg_config.get(ConfigKey.REG_TABLE.value)
        if self.registry_table is None:
            self.logger.error(
                "SnowRegistry._init_config(): Configuration doesn't have "\
                f"registry table defined."
            )


    def update_model_registry(
        self,
        pipeline: Pipeline,
        step: Step,
        ref_id: str = "",
        model_name: str = "",
        model_type: str = "",
        model_path: str = "",
        lib_type: LibType = LibType.Default,
        model: Any = None,
        model_version: str = "",
        user_id: str = ""
    ):
        nmsp: str = self.context.data.get("app_namespace", "")
        app_nm: str = self.context.data.get("app_name", "")
        ts: str = datetime.now().strftime("%Y-%m-%d %H:%M:%S")

        if not ref_id:
            ref_id = model_name.strip().replace(" ", "_")
        ref_id = ref_id.lower()
        mdl_id = SnowRegistry.to_model_id(ref_id)

        mdl_b64, tmp_f = "", "", ""
        if lib_type == LibType.Pytorch:
            mdl_b64, tmp_f = SnowRegistry.to_string_from_model_path(\
                    model_path, mdl_id)
        else:
            mdl_b64 = SnowRegistry.to_string_from_model(model)

        js_params = json.dumps(self.context.params)
        js_metrics = json.dumps(self.context.metrics)

        mdl_ver: str = model_version
        if (not model_version) and self.app_config:
            mdl_ver = self.app_config.version

        column_names = [
            "APP_NAMESPACE", "APP_NAME", "PIPELINE",
            "STEP", "REF_ID", "MODEL_NAME", "MODEL_ID", 
            "MODEL_TYPE",  "MODEL_PATH", "LIB_TYPE", 
            "MODEL_BIN", "STEP_PARAMS", "STEP_METRICS", 
            "USER", "VERSION", "TIMESTAMP"
        ]
    
        session = self.data_connect.create_service_session()

        if not user_id:
            user_id = session.get_current_role()

        df_mdl = pd.DataFrame([[
                nmsp, app_nm, pipeline.pipeline_key, 
                step.func.__name__, ref_id, model_name, mdl_id,
                model_type, model_path, lib_type.value, 
                mdl_b64, js_params, js_metrics,
                user_id, mdl_ver, ts
            ]],
            columns=column_names)

        tbl_nm = AppConfig.T_REG_TBL_DEF
        if self.registry_table is not None:
            tbl_nm = self.registry_table

        params: Dict = self.data_connect.connect_params
        db_nm = params.get(ConfigKey.DB.value, "")
        scm_nm = params.get(ConfigKey.SCH.value)
        session.write_pandas(
            df = df_mdl, 
            database = db_nm,
            schema = scm_nm,
            table_name = tbl_nm, 
            auto_create_table=True,
            overwrite=False
        )

        self.logger.debug(
            "SnowRegistry.update_model_registry(): Save Model "\
            f"[{model_name}] to Table [{tbl_nm}] [OK]."
        )
        # clean up 
        SnowRegistry.clean_up_model_file(tmp_f, model_path)


    def query_model_registry(
            self,
            pipeline: Pipeline,
            step: Step,
            model_name: str = "",
            ref_id: str = "",
            timestamp: str = ""
    ) -> Tuple[Any, str, Dict]:
        nmsp: str = self.context.data.get("app_namespace", "")
        app_nm: str = self.context.data.get("app_name", "")
        tbl_nm: str = self.registry_table if self.registry_table \
                else AppConfig.T_REG_TBL_DEF
        
        session = self.data_connect.create_service_session()
        sdf_mdl: SDF = DFF.create_sdf(tbl_nm, session)

        window_spec = W.partitionBy("APP_NAMESPACE", "APP_NAME", 
                "PIPELINE", "STEP", "MODEL_NAME", "REF_ID")

        sdf_mdl = sdf_mdl.withColumn("max_timestamp", F.max("TIMESTAMP")\
                .over(window_spec))

        sdf_mdl = sdf_mdl.filter(
            (col("APP_NAMESPACE") == nmsp) & 
            (col("APP_NAME") == app_nm) & 
            (col("PIPELINE") == pipeline.pipeline_key) & 
            (col("STEP") == step.func.__name__) & 
            (col("MODEL_NAME") == model_name) & 
            (col("REF_ID") == ref_id) & 
            (col("TIMESTAMP") == col("max_timestamp")) if not timestamp else\
            (col("TIMESTAMP") == timestamp)
        )

        r_mdl_res = sdf_mdl.select("MODEL_BIN", "MODEL_PATH", 
                "STEP_METRICS", "LIB_TYPE").collect()
        
        if len(r_mdl_res) == 0:
            self.logger.warn(
                "SnowRegistry.query_model_registry(): Model"\
                f"[{model_name}], Ref[{ref_id}], TS[{timestamp}] "\
                "query result is empty!"
            )
            return (None, "", {})

        mdl_b64 = r_mdl_res[0]["MODEL_BIN"]
        mdl_path: str = r_mdl_res[0]["MODEL_PATH"]
        js_metrics = r_mdl_res[0]["STEP_METRICS"]
        d_metrics: Dict = json.loads(js_metrics)

        if r_mdl_res[0]["LIB_TYPE"] == LibType.Pytorch.value:
            mdl = SnowRegistry.to_model_file_from_string(
                    mdl_b64, mdl_path)
        else:
            mdl_b64 = r_mdl_res[0]["MODEL_BIN"]
            pkl_mdl = base64.b64decode(mdl_b64)
            mdl = pickle.loads(pkl_mdl)

        return (mdl, mdl_path, d_metrics)


    @staticmethod
    def to_model_id(model_ref_id: str) -> str:
        return hashlib.sha256(model_ref_id.encode("utf-8"))\
                .hexdigest()    


    @staticmethod
    def to_string_from_model(model: Any) -> str:
        """
        convert ML model in python (generally from SKLearn) to B64 
        encoded string.
        """
        pkl_mdl_b64 = ""
        if model is not None: 
            pkl_mdl_bin = pickle.dumps(model)
            pkl_mdl_b64 = base64.b64encode(pkl_mdl_bin).decode()
        
        return pkl_mdl_b64


    @staticmethod
    def to_string_from_model_path(
            model_path: str,
            model_id: str = ""
    ) -> Tuple[str, str]:
        """
        Packege model path and files from file system to B64 encoded
        string and include a temporary file for the compressed file dir.
        """
        s_rs : str = ""
        temp_file : str = model_id + ".zip" if model_id else "model.zip"
        if model_path:
            shutil.make_archive(temp_file, 'zip', model_path)
            with open(temp_file, 'rb') as file:
                s_rs = base64.b64encode(file.read()).decode()

        return (s_rs, temp_file)
    

    @staticmethod
    def to_model_file_from_string(
            model_str: str,
            model_path: str,
            model_id: str = ""
    ) -> Any:
        decoded_mdl = base64.b64decode(model_str)

        temp_file = model_id + ".zip" if model_id else "model.zip"
        with open(temp_file, 'wb') as file:
            file.write(decoded_mdl)

        with zipfile.ZipFile(temp_file, 'r') as zip_ref:
            zip_ref.extractall(model_path)

        os.remove(temp_file)


    @staticmethod
    def clean_up_model_file(
            temp_file: str,
            dir_path: str = ""
    ):
        if temp_file is not None and temp_file:
            os.remove(temp_file)

        if dir_path:
            shutil.rmtree(dir_path)
    


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains AppConfig class representing application
configrations.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


from enum import Enum
import os
import sys
from os.path import exists
from pathlib import Path
from types import ModuleType
from typing import Optional, Union, Dict, Tuple, Any
import logging
import toml
import importlib
from importlib.resources import read_text



class ConfigType(Enum):
    Apps = "apps"
    AppConnects = "app_connects"
    DataSetups = "data_setups"
    BaseApps = "base_apps"
    Streamlits = "streamlit_apps"
    MLPipelines = "ml_pipelines"
    MLOps = "ml_ops"
    OAuthConnects = "oauth_connects"
    DataConnects = "data_connects"
    FeatureStores = "feature_stores"
    ModelRegistries = "model_registries"
    PipelineFlows = "pipeline_flows"



class AppType(Enum):
    Default = "default"
    Streamlit = "streamlit"
    Notebook = "notebook"
    Console = "console"



class ConfigKey(Enum):
    APP_NAME = "app_name"
    NAME = "name"
    APP_SHORT_NAME = "app_short_name"
    TYPE = "type"
    VERSION = "version"
    DOMAIN_ENV = "domain_env"
    AUTH_TYPE = "auth_type"
    INIT_LIST = "init_list"
    ROOT_PATH = "root_path"
    APP_PATH = "app_path"
    SCRIPT = "script"
    SCRIPT_HOME = "script_home"
    CONN_OAUTH = "oauth_connect"
    CONN_DATA = "data_connect"
    STEP_TASKS = "step_tasks"
    RUN = "run"
    INIT = "init"
    CLEAN_UP = "clean_up"
    ACCT = "account"
    DB = "database"
    WH = "warehouse"
    SCH = "schema"
    ROLE = "role"
    USER = "user"
    REG_TABLE = "registry_table"
    EXE_MODE = "execution_mode"
    CLASS_NAME = "class"
    FEAT_STORE = "feature_store"
    MODEL_REG = "model_registry"



class AppConfig:
    """
    This class represents the overall enterprise ai/ml application 
    configurations from a directory of configruation files and its
    instance represents a specific application configuration.

    The bootstrapping priority is the following: 

    1) input custom directory 
    2) .snowflake_ai/conf subdir under current directory 
    3) .snowflake_ai/conf subdir under user_home directory
    4) current directory
    5) home directory
    6) conf subdir under snowflake_ai library installation root dir

    Assuming there is a configruation of an application app_1 under a 
    default group named as group_def:

        >>> from snowflake_ai.common import AppConfig
        ... 
        ... # initialize application config for app_1 under group_def
        >>> ac = AppConfig("group_def.app_1")
        ...
        ... # optionally the configs may be loaded from your custom dir
        ... # ac = AppConfig("group_def.app_1", "custom dir", 
        ... #         "config.toml")
        ...
        ... # to create an actual application, e.g., notebook app
        >>> app = NotebookApp("business_group.app_1")
    """

    DEF_CONF_LIB_PATH = "snowflake_ai.conf"
    DEF_CONF_FILE = "app_config.toml"
    DEF_CONF_DIR = "./.snowflake_ai/conf/"
    DEF_CONN = "snflk_svc_def"
    DEF_CURR_PATH = "."
    DEF_HOME_PATH = "~"    

    # value types for the config keys
    T_DEFAULT = "default"
    T_DEF = "_def"
    T_CONN_SNFLK = "snowflake"
    T_CONN_FILE = "file"
    T_OAUTH = "oauth"
    T_OAUTH_DEVICE = "device_code"
    T_OAUTH_CODE = "auth_code"
    T_OAUTH_CREDS = "client_credentials"
    T_AUTH_SNFLK = "snowflake"
    T_AUTH_KEYPAIR = "keypair"
    T_AUTH_EXT_BROWSER = "externalbrowser"
    T_AUTH_OAUTH = "oauth"    
    T_EXE_SEQ = "sequential"
    T_EXE_PAR = "parallelable"
    T_REG_TBL_DEF = "default_model_registry"


    logging.basicConfig(
        level=logging.ERROR,
        format='%(asctime)s [%(levelname)s]  %(message)s'
    )
    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    # only init one in runtime
    _initialized = False
    _configs_root = ''
    _configs_path = ''

    # store all configs at class level
    _configs: Dict[str, Dict] = {}

    # store appconfig ref by specific app key at class level
    _apps = {}

    # init lib load system path
    _init_lib_path = False


    def __init__(
        self,
        app_key : str,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ):
        """
        Creat app specific AppConfig object.

        Args:
            app_key: A string representing the application; it can have 
                the format of <app_group>.<application> ('<', '>' not 
                included)
            config_dir: directory path string for custom config load
            config_file: file name string for custom config load
        """
        self.logger = AppConfig._logger
        self.app_key = app_key.strip().lower()

        # init all configs and configs dict
        self.root_path = AppConfig._init_all_configs(config_dir, config_file)

        # get group_key.app_name , [apps.<group_key>.<app_name>] app config
        self.app_key, self.app_config = AppConfig.load_app_config(
                self.app_key)
        self.app_name = self.app_config.get(ConfigKey.APP_NAME.value,
                self.app_key.lower().replace('.', '_') )
        self.app_group = AppConfig.split_group_key(self.app_key)[0]
        self.app_short_name = self.app_config.get(
                ConfigKey.APP_SHORT_NAME.value, 
                self.app_key.lower().replace('.', '_')
            )

        # init config dict, set app global attrs, app type, and root path
        self.type =  AppConfig._configs[ConfigType.Apps.value].get(
                ConfigKey.TYPE.value, AppType.Default.value)
        self.root_path =  AppConfig._configs[ConfigType.Apps.value].get(
                ConfigKey.ROOT_PATH.value, 
                os.path.abspath(AppConfig.DEF_CURR_PATH))

        # initialize additional app specific attrs, e.g., name, type, path
        self._init_app_base_config()
        if bool(self.app_key) and bool(self.app_config):
            self._apps[self.app_key] = self

        self.logger.debug(
                f"AppConfig.init(): App_name [{self.app_name}]; "\
                f"App_group [{self.app_group}]; Root_path [{self.root_path}]"\
                f"; Script_home [{self.script_home}]") 
        
        # get base app config
        _, self.app_base_config = AppConfig.get_group_item_config(
                self.app_key, ConfigType.BaseApps.value, self._configs)
        self.logger.debug(
                f"AppConfig.init(): BaseAppConfigs => "\
                f"{self.app_base_config}.")        
        self.app_connect_refs = self.app_base_config.get(
                ConfigType.AppConnects.value, [])
        self.ml_ops_refs = self.app_base_config.get(
                ConfigType.MLOps.value, [])
        self.data_setup_refs = self.app_base_config.get(
                ConfigType.DataSetups.value, [])        
        self.ml_pipeline_refs = self.app_base_config.get(
                ConfigType.MLPipelines.value, [])
        self.logger.debug(f"AppConfig.init(): App_key [{self.app_key}]; "\
                f"App_connect_ref [{self.app_connect_refs}]; "\
                f"Data_setup_ref [{self.data_setup_refs}]")

        # get oauth and data connect configs groups
        d_conn: Dict = AppConfig._configs.get(
                ConfigType.AppConnects.value)
        self.oauth_connect_configs: Dict = {}
        if d_conn is not None:
            d_oauth_conn = d_conn.get(ConfigType.OAuthConnects.value)
            self.oauth_connect_configs = d_oauth_conn
        self.data_connect_configs: Dict  = {}
        if d_conn is not None:
            d_data_conn = d_conn.get(ConfigType.DataConnects.value)
            self.data_connect_configs = d_data_conn
        


    @staticmethod
    def load_configs(
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ) -> Tuple[str, str, Dict]:
        """
        Load configurations from input configuration directory and file if 
        supplied, or load them based on bootstrapping priorities.

        Args:
            config_dir (str): Configuration files directory.
            config_file (str): Configuration file name. If it is none or 
                empty, merging all toml files to create a complete config
                dictionary.

        Returns:
            Tuple[str, str, Dict]: custom config root dir path, dir path
                from where the configurations are loaded, and the dict
                of all loaded configurations (merged all configs).

        Raises:
            ValueError: If config root, config file or path doesn't exist.
        """
        config_rt = None
        if (config_dir is None) or (not config_dir):
            config_rt = AppConfig.DEF_CURR_PATH
            config_dir = AppConfig.DEF_CONF_DIR
        configs = AppConfig._load_toml_files(config_dir, config_file)
        if config_rt != AppConfig.DEF_CURR_PATH:
            config_rt = config_dir
        if config_rt:
            config_rt = os.path.abspath(config_rt)
        AppConfig._logger.debug(
            f"AppConfig.load_configs(): Load from input default directory"\
            f" - Current_root_dir [{config_rt}]; Config_dir [{config_dir}]."
        )

        home = Path.home()
        if not configs:
            config_rt = os.path.abspath(home)
            config_dir = os.path.join(home, AppConfig.DEF_CONF_DIR)
            configs = AppConfig._load_toml_files(config_dir, config_file)
            AppConfig._logger.debug(
                f"AppConfig.load_configs(): Load from home default directory"\
                f" - Home [{home}]; Config_root [{config_rt}]; "\
                f"Config_dir [{config_dir}]."
            )
        
        if not configs:
            config_rt = os.path.abspath(AppConfig.DEF_CURR_PATH)
            config_dir = config_rt
            configs = AppConfig._load_toml_files(config_dir, config_file)
            AppConfig._logger.debug(
                f"AppConfig.load_configs(): Load from input directory "\
                f"after home default check - Config_root [{config_rt}];"\
                f" Config_dir [{config_dir}]."
            )
        
        if not configs:
            config_rt = os.path.abspath(home)
            config_dir = config_rt
            configs = AppConfig._load_toml_files(config_dir, config_file)        
            AppConfig._logger.debug(
                f"AppConfig.load_configs(): Load from home directory "\
                f"directly - Home [{home}]; Config_root [{config_rt}];"\
                f" Config_dir [{config_dir}]."
            )
        
        if not configs:
            config_dir = AppConfig.DEF_CONF_LIB_PATH
            configs = AppConfig.load_default_configs()
            AppConfig._logger.debug(
                f"AppConfig.load_configs(): Load from library default "\
                f"directory - Config_root [{config_rt}];"\
                f" Config_dir [{config_dir}]."
            )
        
        if (config_rt is not None) and config_rt:
            config_rt =  os.path.abspath(config_rt)
            AppConfig._logger.debug(
                f"AppConfig.load_configs(): Output current configuration"\
                f" bootstrap path - Config_root [{config_rt}];"\
                f" Config_dir [{config_dir}]."
            )
        
        if (config_rt is None) or (not exists(config_rt)):
            s = f"AppConfig.load_configs(): Error - config_root "\
                    f"[{config_rt}] doesn't exist!"
            AppConfig._logger.error(s)
            raise ValueError(s)

        return (config_rt, config_dir, configs)


    @staticmethod
    def _load_toml_files(
        config_dir : str, 
        config_file : Optional[str] = None
    ) -> Dict:
        rd = {}
        files_tsd = {}
        toml_files = [] if config_file is None else [config_file]

        if (config_dir is not None) and (config_dir.strip()):
            config_dir = os.path.abspath(config_dir.strip())

        if not exists(config_dir):
            AppConfig._logger.warning(
                f"AppConfig._load_toml_files(): Directory [{config_dir}]"\
                " doesn't exist!"
            )
            return rd

        if (config_file is not None) and bool(config_file.strip()):
            config_file_path = os.path.join(config_dir, config_file)
            with open(config_file_path, 'r') as f:
                try:
                    rd = toml.load(f)
                except Exception as e:
                    raise ValueError(
                        f"AppConfig._load_toml_files(): Cannot load "\
                        f"file [{config_file_path}], check format! "\
                        f"Error - {e}!"
                    )
            return rd
        else:
            toml_files = [
                f for f in os.listdir(config_dir) \
                    if f.lower().endswith(".toml")
            ]
            for toml_file in toml_files:
                config_file_path = os.path.join(config_dir, toml_file)
                with open(config_file_path, 'r') as f:
                    try:
                        toml_dict = toml.load(f)
                    except Exception as e:
                        raise ValueError(
                            f"AppConfig._load_toml_files(): Cannot load "\
                            f"list of files from [{config_file_path}], "\
                            f" check format! Error - {e}!"
                        )
                file_ts = os.path.getmtime(config_file_path)
                config_file = config_file_path

                for key, value in toml_dict.items():
                    key = key.strip().lower()
                    if key not in rd or file_ts > files_tsd[key]:
                        rd[key] = value
                        files_tsd[key] = file_ts

        AppConfig._logger.info(
            f"DataConnect._load_toml_files(): Loaded configuration from "\
            f"directory [{config_dir}]; Loaded files => {toml_files};"\
            f" Configuration keys => {rd.keys()}."
        )
        AppConfig._logger.debug(
            f"DataConnect._load_toml_files(): Loaded configuration from "\
            f"path [{config_dir}]; Detailed App_config [{rd}]."
        )
        return rd


    @staticmethod
    def load_default_configs() -> Dict:
        """
        Load default configuration from libary.

        Returns:
            dict: loaded configuration as a dictionary
        """
        config_file = read_text(
            AppConfig.DEF_CONF_LIB_PATH, 
            AppConfig.DEF_CONF_FILE
        )   
        configs = toml.loads(config_file)
        AppConfig._logger.debug(
            f"AppConfig.load_default_configs(): Loaded default app"\
            f"configuration; Config keys [{configs.keys()}]."
        )
        return configs
    

    @staticmethod
    def get_group_item_config(
        group_item_key: str,
        root_key: str,
        configs: Optional[Union[Dict, None]] = None
    ) -> Tuple[str, Dict]:
        """
        Get sub configuration (dictionary obj) of the section from three level
        of the overall configuration <root>.<group>.<item>.

        Args:
            group_item_key (str): string in form of <group_key>.<item_key>.
            root_key (str): top level configuration section key, e.g., 
                "apps", "app_connects"
            configs (Dict): overall configuration dictionary

        Returns:
            Tuple[str, dict]: tuple of the key string matched in
            a form of group_key.item_key and the dictionary of loaded
            configuration section corresponding to the group_item_key.
        """
        root_key = root_key.strip().lower() if root_key \
                else ConfigType.Apps.value
        if configs is not None and \
                configs.get(root_key) is None:
            s = f"AppConfig.get_group_item_config(): Error "\
                    f"- [{root_key}] is missing!"
            AppConfig._logger.error(s)
            raise ValueError(s)

        if configs is None:
            configs = AppConfig.get_all_configs()

        rd = {}
        gp_itm_key = group_item_key.strip().lower()
        gk, ik = AppConfig.split_group_key(gp_itm_key)
        AppConfig._logger.debug(
            f"AppConfig.get_group_item_config(): Split key - "\
            f"GroupKey [{gk}]; ItemKey[{ik}]."
        )
        if (not gk) and ik:
            k, rd = AppConfig.search_key_by_group(
                ik, root_key, configs
            )
        elif (not gk) and (not ik):
            k, rd = '', {}
        elif gk and (not ik):
            k, rd = f"{gk}.", AppConfig.filter_group_key(
                gk, root_key, configs
            )
        else:
            gs: Dict[str, Dict] = dict(configs[root_key]).get(gk)
            if gs is not None:
                k = f"{gk}.{ik}"
                rd =  gs[ik] if gs.get(ik) is not None else {}
                AppConfig._logger.debug(
                    f"AppConfig.get_group_item_config(): Root_key"\
                    f"[{root_key}]; GroupKey [{gk}]; "\
                    f"ItemConfigs [{gs}]; ItemKey[{ik}]; Config => {rd}."
                )
            else:
                k, rd =  f"{gk}.{ik}", {}

        AppConfig._logger.debug(
            f"AppConfig.get_group_item_config(): Loaded group item "\
            f"configration with key [{k}]; Config => {rd}."
        )
        return (k, rd)


    @staticmethod
    def load_app_config(
        app_key: str, 
        configs: Optional[Union[Dict, None]] = None
    ) -> Tuple[str, Dict]:
        """
        Load app configuration from overall configurations, i.e.
        [apps.<group_key>.<app_name>] config section.

        Args:
            app_key (str): application key.
            configs (Dict): overall configuration dictionary

        Returns:
            Tuple[str, dict]: tuple of the app key string matched in
            a form of group_key.app_name and the dictionary of loaded
            application specific configurations
        """
        return AppConfig.get_group_item_config(
                app_key,
                ConfigType.Apps.value,
                configs
            )


    @staticmethod
    def filter_group_key(group_key: str, top_key:str, configs: Dict) \
        -> Dict:
        """
        Filter group section of configuration dictionary which has the
        structure of top_key: { group_key : { key1: "value1", key2: {} } }
        so that the returned dictionary group section doesn't have any
        nested dictionaries or subsections, i.e., {key1: "value1"} is
        returned as an example.

        Args:
            group_key (str): A string representing the group key
            top_key (str): A string that is the key of parent of group
                sections as the top level.
            configs (Dict): Configuration dictionary

        Returns:
            A dict containing group section's simple key and value pairs
            with any subsections filtered out.
        """
        if top_key not in configs or group_key not in configs[top_key]:
            return {}

        group_dict: dict = configs[top_key][group_key]

        filtered_dict = {k: v for k, v in group_dict.items() \
            if not isinstance(v, dict)}

        return filtered_dict


    @staticmethod
    def search_key_by_group(key: str, top_key:str, configs: Dict) \
        -> Tuple[str, Dict]:
        """
        Search sections of configs dictionary which has the structure of 
        top_key: { group_key: { key: {} } }, and return a tuple as 
        (group_key.key, dict of matched key section).

        Args:
            key (str): A string representing the input key to be searched.
            top_key (str): A string that is the key of parent of group
                sections as the top level.
            configs (Dict): Configuration dictionary

        Returns:
            A tuple containing the group_key.key and dict of matched
            key section.
        """
        if top_key not in configs:
            return '',  {}

        default_match = '',  {}
        first_match = '',  {}

        config_groups :dict = configs[top_key]
        for group_key in sorted(config_groups.keys()):
            group_dict = config_groups[group_key]
            if key in group_dict:
                if (AppConfig.T_DEF in group_key) or \
                        (AppConfig.T_DEFAULT in group_key):
                    default_match = f"{group_key}.{key}", group_dict[key]
                elif first_match == ('', {}):
                    first_match = f"{group_key}.{key}", group_dict[key]

        if default_match != ('', {}):
            return default_match
        else:
            return first_match


    @staticmethod
    def split_group_key(key: str) -> Tuple[str, str]:
        """
        Split an input key by '.' and return group_key and the key without
        group prefix.

        Args:
            key: A string representing the input key to be split.

        Returns:
            A tuple containing the group_key and key without group prefix
            as strings.
        """
        if key is None:
            return '', ''
        key = key.strip().lower()
        split_keys = key.split('.')
        if len(split_keys) > 1:
            group_k = split_keys[0]
            k = split_keys[1]
        else:
            k = split_keys[0]
            group_k = ''
        return (group_k, k)


    @staticmethod
    def load_module(
            script_name: str,
            root_path: str = "",
            script_home: str = ""
    ) -> ModuleType:
        if (script_name) is None or (not script_name):
            AppConfig._logger.warn(
                f"AppConfig.load_module(): Module script name is "\
                f"[{script_name}]!"
            )
            return None
        
        if (script_name is not None) and ('.' in script_name):
            script_name = script_name.split('.')[0]
        
        if (not AppConfig._init_lib_path) and root_path and script_home:
            p = os.path.join(root_path, script_home)
            sys.path.insert(0, os.path.expanduser(p))
            AppConfig._init_lib_path = True

        AppConfig._logger.debug(
            f"AppConfig.load_module(): Load [{script_name}] "\
            f"from root_path[{root_path}]/script_home[{script_home}]."
        )
        module = None
        try:
            module = importlib.import_module(script_name)
            AppConfig._logger.debug(
                f"AppConfig.load_module(): Module [{script_name}] loaded!"
            )
        except Exception as e:
            AppConfig._logger.error(
                f"AppConfig.load_module(): Error - {e} in loading module "\
                f"[{script_name}]!"
            )
        return module
    

    @staticmethod
    def get_qualified_key(root_key: str, key: str) -> str:
        """
        Get fully qualified key in form of <group_key>.<key> by searching
        key from root_key.

        Returns:
            A string representing a fully qualified key in config dict.
        """
        if root_key not in ConfigType._value2member_map_:
            raise ValueError(
                f"AppConfig.get_qualified_key(): Error - [{key}] not found!"
            )
        
        gk, k = AppConfig.split_group_key(key)
        if k:
            configs: dict = AppConfig.get_all_configs()
            ret_k, _ = AppConfig.search_key_by_group(k, root_key, configs)
            if not ret_k:
                raise ValueError(
                    f"AppConfig.get_qualified_key(): Error - Key Empty!"
                )
            else:
                g_dict = dict(configs[root_key]).get(gk)
                if g_dict is not None:
                    ret_k = f"{gk}.{k}"
        else:
            raise ValueError(
                f"AppConfig.get_qualified_key(): Error with key [{key}]!"
            )
        return ret_k



    def _init_app_base_config(self):
        # get app long name and group config
        self.name = self.app_config.get(ConfigKey.NAME.value)
        self.group_config = AppConfig.filter_group_key(
            self.app_group, ConfigType.Apps.value, self.get_all_configs()
        )

        s = self.app_config.get(ConfigKey.TYPE.value)
        if s is not None and s:
            self.type = s
        elif self.type is None or not self.type:
            self.type = AppType.Default.value
        self.version = self.app_config.get(ConfigKey.VERSION.value)
        self.domain_env = self.app_config.get(ConfigKey.DOMAIN_ENV.value)

        # app and root path (for app deployment)
        self.app_path = self.app_config.get(ConfigKey.APP_PATH.value, 
                AppConfig.DEF_CURR_PATH)
        s = self.app_config.get(ConfigKey.ROOT_PATH.value)
        if s is not None and s:
            self.root_path = s
        if self.root_path is None or not self.root_path:
            self.root_path = ""

        # set app script home path dir (for all app related scripts)
        self.script_home = self.app_config.get(ConfigKey.SCRIPT_HOME.value,
                 AppConfig.DEF_CURR_PATH)


    @classmethod
    def _init_all_configs(
        cls,
        config_dir : Optional[Union[str, None]] = None, 
        config_file : Optional[Union[str, None]] = None
    ) -> str:
        """
        Initialize the dictionary of all application configurations

        Returns:
            The root directory path where the configs are loaded from.
        """        
        if not cls._initialized:
            cls._configs_root, cls._configs_path, cls._configs = \
                AppConfig.load_configs(config_dir, config_file)
            cls._initialized = True
        return cls._configs_root

    
    @classmethod
    def get_configs_root(cls) -> str:
        """
        Get root path of directory where configurations are loaded.

        Returns:
            str: root path string of directory containing all 
                configuration files.
        """
        if not cls._configs_root:
            cls._configs_root = cls._init_all_configs()
        return cls._configs_root
    

    @classmethod
    def get_configs_path(cls) -> str:
        """
        Get path of directory where configurations are loaded.

        Returns:
            str: path string of directory containing all 
                configuration files.
        """
        if not cls._configs_path:
            cls._init_all_configs()
        return cls._configs_path
    

    @classmethod
    def get_all_configs(cls) -> Dict[str, Dict]:
        """
        Get overall configurations.

        Returns:
            dict: all configurations.
        """ 
        if not cls._configs:
            cls._init_all_configs()
        return cls._configs


    @classmethod
    def get_all_apps(cls) -> Dict:
        """
        Get all initialized applications.

        Returns:
            dict: all initialized applications.
        """  
        return cls._apps


    @property
    def apps(self) -> Dict:
        """
        Get/set a dictionary of all initialized applications.

        Returns:
            dict: a dictionary of all initialized applications.
        """   
        return AppConfig._apps


    def get_app_type(self) -> str:
        """
        Get application configured type. Currently supported type
        is "streamlit_ml"

        Returns:
            str: application type string.
        """   
        return self.type


    def get_root_path(self) -> str:
        """
        Get overall apps configuration root path

        Returns:
            str: root directory path for all applications
        """   
        return self.root_path


    def get_app_path(self) -> str:
        """
        Get the directory of the specified application

        Returns:
            str: the directory of the specified application
        """   
        return self.app_path
    

    def get_app_namespaces(self) -> Tuple[str, str]:
        """
        Get application key in form of (group_key, app_name)

        Returns:
            tuple: (group_key, app_name)
        """   
        rs: Tuple[str, str] = ("" ,"")
        if self.app_key:
            s = self.app_key.rsplit('.', 1)
            if len(s) == 1:
                rs = ("", s[0])
            elif len(s) > 1:
                rs = (s[0], s[1])
        return rs
    


from .file_connect import *
from .authcode_connect import *
from .devicecode_connect import *
from .clientcreds_connect import *
from .snow_connect import *
from .dataframe_factory import *
from .connect_manager import *



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains an utility class of SnowTransform
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


from typing import Optional, Union, Tuple, List
from datetime import *
from dateutil.relativedelta import *
import snowflake.snowpark.functions as F
from snowflake.snowpark import Session
from snowflake.snowpark import Window as W
from snowflake.snowpark import DataFrame as SDF



class SnowTransform:
    """
    This class contains a set of static methods for transforming Snowflake
    dataframe including creating columns and tables for ease of time series
    data analysis.
    """

    @staticmethod
    def create_year_column(
        sdf: SDF, 
        dt_col_nm : str,
        yr_col_nm: Optional[str] = 'YEAR'
    ) -> SDF:
        """
        Create a new Snowflake dataframe column (named as yr_col_nm) for the 
        calendar year number (e.g., 2023, etc.) from the existing Snowflake 
        dataframe based on input date column (named as dt_col_nm)

        Args:
            sdf (Snowflake Dataframe): input Dataframe
            dt_col_nm (str): existing date column name, e.g. 'TIME_IDX_DT'
            yr_col_nm (str): year column name, default to 'YEAR'; its value type 
                is int.

        Returns:
            Snowflake Dataframe: Dataframe with new year column (int) added.
        """
        if str.strip(dt_col_nm) == '':
            dt_col_nm = "TIME_IDX_DT"
        if str.strip(yr_col_nm) == '':
            yr_col_nm = "YEAR"

        return sdf.withColumn(
            yr_col_nm, 
            F.year(dt_col_nm)
            )
    

    @staticmethod
    def create_week_column(
        sdf: SDF, 
        dt_col_nm : str,
        wk_col_nm: Optional[str] = 'WEEK'
    ) -> SDF:
        """
        Create a new Snowflake dataframe column (named as wk_col_nm) for the
        week of year string formatted as '00' padded with '0' (e.g., '01', etc.)
        from the existing Snowflake dataframe based on input date column 
        (named as dt_col_nm).
        NOTE: Not applying additional filters on the resulted dataframe as 
        the additional filtering may change week partition behavior.

        Args:
            sdf (Snowflake Dataframe): input Dataframe
            dt_col_nm (str): existing date column name, e.g. 'TIME_IDX_DT'
            wk_col_nm (str): week column name, default to 'WEEK'; its value 
                is formatted as '00' string

        Returns:
            Snowflake Dataframe: Dataframe with a new week column (str) added.
        """
        if str.strip(dt_col_nm) == '':
            dt_col_nm = "TIME_IDX_DT"
        if str.strip(wk_col_nm) == '':
            wk_col_nm= "WEEK"

        sdf = sdf.withColumn(
            wk_col_nm, 
            F.when(
                (F.month(dt_col_nm) == 1) & (F.weekofyear(dt_col_nm) >= 52),
                F.trim(F.to_char(F.lit(1), "00"))
            )
            .otherwise(
                F.trim(F.lpad(F.weekofyear(dt_col_nm), 2, F.lit('0')))
            )
        )

        window = W.partition_by(wk_col_nm).order_by(dt_col_nm)
        sdf = sdf.withColumn(
            wk_col_nm, 
            F.when(
                F.dayofweek(dt_col_nm) == 1,
                sdf[wk_col_nm]
            )
            .otherwise(
                F.last_value(sdf[wk_col_nm], ignore_nulls=True).over(window)
            )
        )

        return sdf
    

    @staticmethod
    def create_year_month_column(
        sdf: SDF, 
        dt_col_nm : str,
        yr_col_nm: Optional[str] = '',
        mo_col_nm: Optional[str] = '',
        mo_str_col_nm: Optional[str] = ''
    ) -> SDF:
        """
        Create two types Snowflake dataframe columns if the input column 
        names are not empty - Column (yr_col_nm) for the year number, and 
        Column (mo_col_nm) for the month number, and Column (mo_str_col_nm)
        for the month string formatted as 'MM padded with '0' (e.g., '01' 
        for Jan, etc.) from the existing Snowflake dataframe based on input 
        date column (named as dt_col_nm).

        Args:
            sdf (Snowflake Dataframe): input Dataframe
            dt_col_nm (str): existing date column name, e.g. 'TIME_IDX_DT'
            yr_col_nm (str): new year column name if not empty, e.g. 'YEAR'; 
                its value type is int.
            mo_col_nm (str): new month column name if not empty, e.g. 'MONTH';
                its value type is int (1 - Jan, 2 - Feb, etc.)
            mo_str_col_nm (str): new month column name, its value type 
                is formatted as 'MM' string ('01' - Jan, '02' - Feb, etc.)

        Returns:
            Snowflake Dataframe: Dataframe with new year column (int) and 
                month column (int) and month string column (int) added.
        """
        if str.strip(dt_col_nm) == '':
            dt_col_nm = "TIME_IDX_DT"

        if str.strip(yr_col_nm) != '':
            sdf = sdf.withColumn(
                yr_col_nm,
                F.year(dt_col_nm)
            )

        if str.strip(mo_col_nm) != '':
            sdf = sdf.withColumn(
                mo_col_nm, 
                F.month(dt_col_nm)                
            )

        if str.strip(mo_str_col_nm) != '':
            sdf = sdf.withColumn(
                mo_str_col_nm,
                F.trim(
                    F.lpad(F.month(dt_col_nm), 2, F.lit('0'))
                )               
            )
        return sdf
    

    @staticmethod
    def create_weekly_time_index_column(
        sdf: SDF, 
        dt_col_nm : str,
        yr_col_nm: Optional[str] = '',
        wk_col_nm: Optional[str] = '',
        time_idx_col_nm: Optional[str] = 'TIME_IDX_WK'
    ) -> SDF:
        """
        Create a new Snowflake dataframe time index column (named as 
        time_idx_col_nm) for week of year string formatted as 'YYYYMM' from
        input snowflake dataframe based on the existing date column 
        (dt_col_nm). Also, if yr_col_nm and wk_col_nm are not
        empty, new year number column and week string column are created.

        Args:
            sdf (Snowflake Dataframe): input Dataframe
            dt_col_nm (str): existing date column name, e.g. 'TIME_IDX_DT'
            yr_col_nm (str): new year column name, e.g. 'YEAR'; its value type
                is int
            wk_col_nm (str): new week column name, e.g. 'WEEK'; its value
                type is string, e.g., '01' is 1st week of the year.
            time_idx_col_nm (str): new weekly time index column name; default
                value to 'TIME_IDX_WK'; its value formatted as 'YYYYWW' string,
                e.g. '202301' for the 1st week of year 2023.

        Returns:
            Snowflake Dataframe: Dataframe with new weekly time index column
                (int), optionally new year column (int), week column (str)
                added.
        """
        if str.strip(dt_col_nm) == '':
            dt_col_nm = "TIME_IDX_DT"
        
        if str.strip(yr_col_nm) != '':
            sdf = SnowTransform.create_year_column(
                sdf, dt_col_nm, yr_col_nm
            )

        if str.strip(wk_col_nm) != '':
            sdf = SnowTransform.create_week_column(
                sdf, dt_col_nm, wk_col_nm
            )        

        return sdf.withColumn(time_idx_col_nm, F.concat(
                F.trim(F.to_char(F.col(yr_col_nm), "0000")), 
                F.trim(F.col(wk_col_nm)))
            )
    

    @staticmethod
    def create_year_week_month_column(
        sdf: SDF, 
        dt_col_nm : str,
        yr_col_nm: Optional[str] = 'YEAR',
        wk_col_nm: Optional[str] = 'WEEK',    
        mon_col_nm: Optional[str] = 'MONTH',
        time_idx_col_nm: Optional[str] = 'TIME_IDX_WK'
    ) -> SDF:
        """
        Create a new Snowflake dataframe time index column (named as 
        time_idx_col_nm) for week of year string formatted as 'YYYYMM' from
        input snowflake dataframe. NOTE: the week and month are calculation
        based on traditional western system and if the week spans two months,
        it is counted for the month that has Thursday of the week. Also, 
        if yr_col_nm, wk_col_nm, mon_col_nm are not empty, new year, month 
        number columns and week string column are created.

        Args:
            sdf (Snowflake Dataframe): input Dataframe.
            dt_col_nm (str): existing date column name, e.g. 'TIME_IDX_DT'
            yr_col_nm (str): new year column name, e.g. 'YEAR'; its value type
                is int.
            wk_col_nm (str): new week column name, e.g. 'WEEK'; its value
                type is string, e.g., '01' is 1st week of the year.
            mon_col_nm (str): new month column name, e.g. 'MONTH'; its value
                type is int.
            time_idx_col_nm (str): new weekly time index column name; default
                value to 'TIME_IDX_WK'; its value formatted as 'YYYYWW' string,
                e.g. '202301' for the 1st week of year 2023.

        Returns:
            Snowflake Dataframe: Dataframe with new weekly time index column
                (int), optionally new year column (int), week column (str),
                month column (int) added.
        """
        sdf = SnowTransform.create_weekly_time_index_column(
            sdf, 
            dt_col_nm=dt_col_nm,
            yr_col_nm=yr_col_nm,
            wk_col_nm=wk_col_nm, 
            time_idx_col_nm=time_idx_col_nm
        )

        if str.strip(mon_col_nm) != '':
            sdf = sdf.withColumn(mon_col_nm, F.month(dt_col_nm))
        
        sdf = sdf.withColumn("THURSDAY_MONTH", F.month(F.date_add(
                dt_col_nm, 4 - F.dayofweek(dt_col_nm))))

        window_wk = W.partitionBy(time_idx_col_nm).orderBy(dt_col_nm)

        sdf = sdf.withColumn("MAJORITY_MONTH", 
                             F.first_value("THURSDAY_MONTH").over(window_wk))

        sdf = sdf.withColumn(
            mon_col_nm,
            F.when(
                (F.col(mon_col_nm) == 1) & (F.col("MAJORITY_MONTH") == 1),
                F.lit(1)
            ).when (
                ((F.col(mon_col_nm) == 12) & (F.col("MAJORITY_MONTH") == 1)),
                F.col(mon_col_nm)
            ).when(
                F.col(mon_col_nm) != F.col("MAJORITY_MONTH"),
                F.col("MAJORITY_MONTH")
            ).otherwise(
                F.col(mon_col_nm)
            )
        )
        sdf = sdf.drop("THURSDAY_MONTH", "MAJORITY_MONTH")
        return sdf


    @staticmethod
    def create_quarter_column(
        sdf: SDF, 
        mon_col_nm: str = '',
        qtr_col_nm: Optional[str] = 'QUARTER'
    ) -> SDF:
        """
        Create a new Snowflake dataframe column for the calendar quarter number 
        from the input Snowflake dataframe based on the existing month
        column (named as mon_col_nm)

        Args:
            sdf (Snowflake Dataframe): input Dataframe
            mon_col_nm (str): existing month column name, e.g. 'MONTH'
            qtr_col_nm (str): quarter column name, default to 'QUARTER'; 
                its value type is int.

        Returns:
            Snowflake Dataframe: Dataframe with a new quarter column (int)
                added based on existing month column.
        """
        if str.strip(mon_col_nm) == '':
            mon_col_nm = "MONTH"
        
        sdf = sdf.withColumn(
            qtr_col_nm, 
            F.when(
                F.col(mon_col_nm).between(1, 3), F.lit(1)
            ).when(
                F.col(mon_col_nm).between(4, 6), F.lit(2)
            ).when(
                F.col(mon_col_nm).between(7, 9), F.lit(3)
            ).otherwise(
                F.lit(4)
            )
        )
        return sdf
    

    @staticmethod
    def create_month_quarter_column(
        sdf: SDF, 
        dt_col_nm : str,
        mon_col_nm: Optional[str] = "MONTH",
        qtr_col_nm: Optional[str] = "QUARTER"
    ) -> SDF:
        """
        Create new Snowflake dataframe columns for the month and quater
        based on the input Snowflake dataframe with the existing date column.

        Args:
            sdf (Snowflake Dataframe): input Dataframe
            mon_col_nm (str): existing month column name, e.g. 'MONTH'
            qtr_col_nm (str): quarter column name, default to 'QUARTER'; 
                its value type is int.

        Returns:
            Snowflake Dataframe: Dataframe with new month column (int)
                using standard calendar system and new quarter column 
                (int) added.
        """
        if str.strip(dt_col_nm) == '':
            dt_col_nm = "TIME_IDX_DT"

        if str.strip(mon_col_nm) != '':
            sdf = sdf.withColumn(mon_col_nm, F.month(dt_col_nm))

        if str.strip(qtr_col_nm) != '':
            sdf = sdf.withColumn(qtr_col_nm, F.quarter(dt_col_nm))

        return sdf
    

    @staticmethod
    def generate_daily_time_index(
        session: Session, 
        start_dt: Union[str, datetime, date],
        end_dt: Union[str, datetime, date],
        time_idx_col_nm: Optional[str] = 'TIME_IDX_DT',
        time_idx_tbl_nm : Optional[str] = ''
    ) -> SDF:
        """
        Generate a daily time index column (named as time_idx_col_nm) with a 
        series of dates formatted as 'YYYY-MM-DD' date objects for a new 
        Snowflake dataframe based on dates from the start_dt to end_dt 
        (inclusive).

        Args:
            session (Session): a Snowflake session
            start_dt (str, datetime, date): starting date object can be parsed
                as 'YYYY-MM-DD' date object.
            end_dt (str, datetime, date): ending date object can be parsed
                as 'YYYY-MM-DD' date object.
            time_idx_col_nm (str): new time index column name; default
                value to 'TIME_IDX_DT', its value is formatted as 'YYYY-MM-DD'.
            time_idx_tbl_nm (str): new dataframe persisted table name; default
                value to '' which is without persisting the dataframe.

        Returns:
            Snowflake Dataframe and Table: Dataframe with a single daily time
                index column (date) typed date object.
        """
        if isinstance(start_dt, datetime):
            start_dt :str = start_dt.strftime("%Y-%m-%d")
        elif isinstance(start_dt, date):
            dt = datetime(
                year=start_dt.year, 
                month=start_dt.month,
                day=start_dt.day,
            )
            start_dt :str = dt.strftime("%Y-%m-%d")
        
        if isinstance(end_dt, datetime):
            end_dt :str = end_dt.strftime("%Y-%m-%d")
        elif isinstance(end_dt, date):
            dt = datetime(
                year=end_dt.year, 
                month=end_dt.month,
                day=end_dt.day,
            )
            end_dt :str = dt.strftime("%Y-%m-%d")

        session.sql("ALTER SESSION SET WEEK_OF_YEAR_POLICY=1, WEEK_START=7")\
            .collect()
        sdf = session.sql(
            f"SELECT '{start_dt}'::DATE + VALUE::INT AS " \
            f"{time_idx_col_nm} FROM TABLE(FLATTEN(ARRAY_GENERATE_RANGE("\
            f"0, DATEDIFF('DAY', '{start_dt}'::DATE, '{end_dt}'::DATE)+1)))"
        )

        if time_idx_tbl_nm:
            sdf = sdf.sort(F.col(time_idx_col_nm).asc())
            sdf.write.mode("overwrite").save_as_table(time_idx_tbl_nm)
            sdf = session.table(time_idx_tbl_nm).to_df(time_idx_col_nm)
        return sdf


    @staticmethod
    def generate_weekly_time_index(
        session: Session, 
        start_dt: Union[str, datetime, date],
        end_dt: Union[str, datetime, date],
        time_idx_col_nm: Optional[str] = 'TIME_IDX_WK',
        time_idx_int_col : Optional[str] = 'TIME_IDX',
        time_idx_tbl_nm : Optional[str] = ''
    ) -> SDF:
        """
        Generate a weekly time index column (named as time_idx_col_nm, 
        default to 'TIME_IDX_WK') with a series of 'YYYYWW' weeks of
        year sequence strings, and its corresponding integer based weekly
        time index column (named as time_idx_int_col, default to 'TIME_IDX') 
        using traditional western system for a new Snowflake dataframe based
        on dates from the start_dt to end_dt (inclusive).
        NOTE: Use this method to create or show week time index dataframe or
        table. It is NOT advised to apply any additional filters which may 
        result in wrong time index.

        Args:
            session (Session): a Snowflake session
            start_dt (str, datetime, date): starting date object can be parsed
                as 'YYYY-MM-DD' date object.
            end_dt (str, datetime, date): ending date object can be parsed
                as 'YYYY-MM-DD' date object.
            time_idx_col_nm (str): new weekly time index column name; default
                value to 'TIME_IDX_WK', its value is type of string and formatted
                as 'YYYYWW', e.g. '202301' - 1st week of year 2023.
            time_idx_int_col (str): new weekly time index column name; default
                value to 'TIME_IDX', its value type is int, e.g., 1, 2, .., etc.
            time_idx_tbl_nm (str): new dataframe persisted table name; default
                value to '' which is without persisting the dataframe.

        Returns:
            Snowflake Dataframe and Table: Dataframe with a weekly time index
                column (str) formatted as 'YYYYWW', and a integer time index
                column (int) named as time_idx_int_col default to 'TIME_IDX', 
                and other dates related columns - year (int), quarter (int),
                month (int), week (str) formatted as '00', week_start_dt 
                column (date), week_end_dt column (date).
        """
        session.sql("ALTER SESSION SET WEEK_OF_YEAR_POLICY=1, WEEK_START=7")\
            .collect()
        
        sdf : SDF = SnowTransform.generate_daily_time_index(
            session, start_dt, end_dt, "TIME_IDX_DT"
        )

        sdf = SnowTransform.create_year_week_month_column(
            sdf, "TIME_IDX_DT", "YEAR", "WEEK", "MONTH", time_idx_col_nm
        )

        sdf = SnowTransform.create_quarter_column(
            sdf, "MONTH", "QUARTER"
        )

        sdf = sdf.with_column(
                "WEEK_DAY", F.dayofweek("TIME_IDX_DT")
            ).filter((F.col("WEEK_DAY") == 1)).select(
                F.col(time_idx_col_nm),
                F.col("TIME_IDX_DT").alias("WEEK_START_DT"),
                F.col("YEAR"), 
                F.col("QUARTER"), 
                F.col("MONTH"), 
                F.col("WEEK")
            )

        sdf = sdf.with_column(
                "WEEK_END_DT", F.date_add("WEEK_START_DT", 6)
            ).distinct().sort(time_idx_col_nm)

        window = W.orderBy(time_idx_col_nm)
        sdf = sdf.withColumn(
            time_idx_int_col, F.row_number().over(window) - 1
        )
        
        sdf = sdf.select(
                F.col(time_idx_int_col),
                F.col(time_idx_col_nm),
                F.col("WEEK_START_DT"),
                F.col("WEEK_END_DT"),
                F.col("YEAR"),
                F.col("QUARTER"), 
                F.col("MONTH"), 
                F.col("WEEK")
            ).sort(time_idx_int_col)
        
        if time_idx_tbl_nm:
            sdf.write.mode("overwrite").save_as_table(time_idx_tbl_nm)
            sdf = session.table(time_idx_tbl_nm).to_df(
                time_idx_int_col, time_idx_col_nm, 
                "WEEK_START_DT", "WEEK_END_DT", 
                "YEAR", "QUARTER", "MONTH", "WEEK"
            )

        return sdf


    @staticmethod
    def generate_week_start_date(
        session: Session,
        start_dt: Union[str, datetime, date],
        end_dt: Union[str, datetime, date],
        day_of_week: int = 1,
        yr_col_nm: Optional[str] = "YEAR",
        wk_col_nm: Optional[str] = "WEEK",
        time_idx_col_nm: Optional[str] = 'TIME_IDX_DT',
        time_idx_tbl_nm : Optional[str] = ''
    ) -> SDF:
        """
        Generate a weekly time index column (named as time_idx_col_nm) with
        its values as the week start date formatted as 'YYYY-MM-DD'
        calculated from day_of_week using traditional western system for a
        new Snowflake dataframe based on dates from the start_dt to end_dt 
        (inclusive).

        Args:
            session (Session): a Snowflake session
            start_dt (str, datetime, date): starting date object can be parsed
                as 'YYYY-MM-DD' date object.
            end_dt (str, datetime, date): ending date object can be parsed
                as 'YYYY-MM-DD' date object.
            day_of_week (int): number indicating which day to start the week;
                in traditional western system, it is default to 1 to indicate Sunday
                as the start of the week.
            yr_col_nm (str): new column name if not empty to get year number.
            wk_col_nm (str): new column name if not empty to get week string.
            time_idx_col_nm (str): new weekly time index column name; default
                value to 'TIME_IDX_WK', its value is type of string and formatted
                as 'YYYYWW', e.g. '202301'.
            time_idx_int_col (str): new weekly time index column name; default
                value to 'TIME_IDX', its value type is int.
            time_idx_tbl_nm (str): new dataframe persisted table name; default
                value to '' which is without persisting the dataframe.

        Returns:
            Snowflake Dataframe and Table: Dataframe with a weekly time index
                using week start date as column (date), year column (int), and 
                week column (str).    
        """
        sdf : SDF = SnowTransform.generate_daily_time_index(
            session, 
            start_dt,
            end_dt,
            time_idx_col_nm,
            '' 
        )
        sdf = SnowTransform.create_year_column(
            sdf, time_idx_col_nm, yr_col_nm
        )
        sdf = SnowTransform.create_week_column(
            sdf, time_idx_col_nm, wk_col_nm
        )

        sdf = sdf.with_column(
                "WEEK_DAY", F.dayofweek(time_idx_col_nm)
            )

        sdf = sdf.filter((F.col("WEEK_DAY") == day_of_week))\
            .select(
                time_idx_col_nm, 
                yr_col_nm, 
                wk_col_nm,
            )

        if time_idx_tbl_nm:
            sdf = sdf.sort(F.col(time_idx_col_nm).asc())
            sdf.write.mode("overwrite").save_as_table(time_idx_tbl_nm)
            sdf = session.table(time_idx_tbl_nm).to_df(time_idx_col_nm)

        return sdf
    

    @staticmethod
    def generate_grouped_time_index(
        sdf: SDF, 
        time_idx_col_nm: str, 
        incl_cols = [], 
        time_idx_int_col : Optional[str] = 'TIME_IDX'
    ) -> SDF:
        """
        Generate a time integer index column (named as time_idx_int_col, 
        default to 'TIME_IDX') based on another ordered time index column
        associated with a group of selected columns.
        NOTE: Use this method to create integer based time index columns with
        selection of a group of columns without additional fitering applied.

        Args:
            sdf (Snowflake Dataframe): input Snowflake dataframe.
            time_idx_col_nm (str): existing time index columns, e.g, 
                TIME_IDX_WK, or TIME_IDX_DT.
            incl_cols (list of string): a list of column names to be selected.
            time_idx_int_col (str): new integer based time index column name; 
                default value to 'TIME_IDX', its value type is int, 
                e.g., 1, 2, .., etc. sequential time index numbers.

        Returns:
            Snowflake Dataframe: Dataframe with added integer based time index
                column (int) together with included columns.
        """   
        if time_idx_col_nm is None or not time_idx_col_nm:
            time_idx_col_nm = "TIME_IDX_NM"

        cols: List[str] = [str(c) for c in incl_cols]
        cols.append(time_idx_col_nm)
        cols_lst = list(set(cols))

        # Convert generator expression to a list of columns
        cols_lst = [str(c) for c in cols_lst]

        sdf = sdf.select(*cols_lst).distinct().sort(time_idx_col_nm)

        window = W.orderBy(time_idx_col_nm)
        sdf = sdf.withColumn(
            time_idx_int_col, F.row_number().over(window) - 1
        )
        return sdf


    @staticmethod
    def generate_monthly_time_index(
        session: Session, 
        start_dt: Union[str, datetime, date],
        end_dt: Union[str, datetime, date],
        yr_col_nm: Optional[str] = '',
        mo_col_nm: Optional[str] = '',
        time_idx_col_nm: Optional[str] = 'TIME_IDX',
        time_idx_tbl_nm : Optional[str] = ''
    ) -> SDF:
        """
        Generate a monthly time index column (named as time_idx_col_nm) 
        based on traditional western system for a new Snowflake dataframe 
        from the start_dt to end_dt (inclusive).
        NOTE: Use this method to create a monthly index dataframe or table.
        It is NOT advised to apply any additional filters which may result in
        wrong time index.

        Args:
            session (Session): a Snowflake session
            start_dt (str, datetime, date): starting date object can be parsed
                as 'YYYY-MM-DD' date object.
            end_dt (str, datetime, date): ending date object can be parsed
                as 'YYYY-MM-DD' date object.
            yr_col_nm (str): new year column name if not empty, e.g. 'YEAR'; 
                its value type is int.
            mo_col_nm (str): new month column name if not empty, e.g. 'MONTH';
                its value type is int (1 - Jan, 2 - Feb, etc.)
            time_idx_col_nm (str): new monthly time index column name; default
                value to 'TIME_IDX', its value type is int.
            time_idx_tbl_nm (str): new dataframe persisted table name; default
                value to '' which is without persisting the dataframe.

        Returns:
            Snowflake Dataframe and Table: Dataframe with a monthly time index
                column (int), year column (int), month column (int).
        """
        sdf : SDF = SnowTransform.generate_daily_time_index(
            session, start_dt, end_dt, "TIME_IDX_DT"
        )

        sdf = SnowTransform.create_year_month_column(
            sdf, "TIME_IDX_DT", yr_col_nm, mo_col_nm
        )

        sdf = sdf.select(yr_col_nm, mo_col_nm)\
                .distinct().sort(yr_col_nm, mo_col_nm)

        window = W.orderBy(yr_col_nm, mo_col_nm)
        sdf = sdf.withColumn(
                time_idx_col_nm, F.row_number().over(window) - 1
            ).select(time_idx_col_nm, yr_col_nm, mo_col_nm)\
            .sort(F.col(time_idx_col_nm).asc(),
                  F.col(yr_col_nm).asc(),
                  F.col(mo_col_nm).asc()
            )
            
        if time_idx_tbl_nm:
            sdf.write.mode("overwrite").save_as_table(time_idx_tbl_nm)
            sdf = session.table(time_idx_tbl_nm).to_df(
                    time_idx_col_nm, yr_col_nm, mo_col_nm)

        return sdf


    @staticmethod
    def get_year_week(
        session: Session,
        dt: Union[str, datetime, date] = None
    ) -> Tuple[str, int, int]:
        """
        Get input date's year and week based on traditional western system.
        
        Return:
            Tuple: a week index string of 'YYYYWW' (str), year number (int)
                and week number (int)
        """
        if dt is None:
            dt : datetime = date.today()
        if isinstance(dt, date):
            dt : datetime = datetime(
                year=dt.year, 
                month=dt.month,
                day=dt.day,
            )
        elif isinstance(dt, str):
            dt: datetime = datetime.strptime(dt, "%Y-%m-%d")

        session.sql("ALTER SESSION SET WEEK_OF_YEAR_POLICY=1, WEEK_START=7")\
            .collect()
        year = dt.year
        sdf : SDF = SnowTransform.generate_daily_time_index(
            session, datetime(year, 1, 1), datetime(year, 12, 31), "DT"
        )
        sdf = sdf.withColumn(
            "WK", 
            F.when(
                (F.month("DT") == 1) & (F.weekofyear("DT") >= 52),
                F.trim(F.to_char(F.lit(1), "00"))
            )
            .otherwise(
                F.trim(F.lpad(F.weekofyear("DT"), 2, F.lit('0')))
            )
        )

        sdf = sdf.withColumn("YR", F.year("DT"))
    
        window = W.partition_by("WK").order_by("DT")
        sdf = sdf.withColumn(
            "WK", 
            F.when(
                F.dayofweek("DT") == 1,
                sdf["WK"]
            )
            .otherwise(
                F.last_value(sdf["WK"], ignore_nulls=True).over(window)
            )
        )
        sdf = sdf.withColumn("YW", F.concat(
            F.trim(F.to_char(F.col("YR"), "0000")), 
            F.trim(F.col("WK")))
        )
        yw_lst = sdf.filter(F.col("DT") == dt).select("YW", "YR", "WK")\
            .collect()
        yw = {}
        if yw_lst:
            yw = yw_lst[0]
        return (str(yw["YW"]), int(yw["YR"]), int(yw["WK"]))


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains OAuthConnect class representing a generic OAuth
connection configuration.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import sys
from typing import Optional, Dict, Tuple, List, Union
import logging
import jwt

from snowflake_ai.common import ConfigKey, ConfigType
from snowflake_ai.common import AppConfig, AppConnect



class OAuthConnect(AppConnect):
    """
    This class represents a generic application connect for 
    OAuth grant flow.

    To use this class, create instance from AppConfig or extend
    it in a child class for creation of a specific OAuth connect
    instance, e.g., AuthCodeConnect as following :

        >>> from snowflake_ai.connect import AuthCodeConnect
        ...
        >>> connect = AuthCodeConnect()
        >>> connect.authorize_request()
        >>> connect.grant_request()
    """

    K_OAUTH_CONN = ConfigType.OAuthConnects.value
    K_INIT_LIST = ConfigKey.INIT_LIST.value
    K_CLIENT_SECRET = "client_secret_env"

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    _initialized = False

    # oauth_connects.<connect_name> : configs dict
    _init_connect_params = {}


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        super().__init__(connect_key, app_config)
        self.logger = OAuthConnect._logger

        self.oauth_connect_key = None

        if self.connect_key:
            _, self.oauth_connect_key = AppConfig.split_group_key(
                self.connect_key
            )
            self.setup_connect(self.connect_params)

        # setup default oauth connect
        if (connect_key is None) or (not connect_key):
            d_configs : Dict = None
            oauth_configs: Dict = None
            if self.app_config is None:
                d_configs = self._configs.get(ConfigType.AppConnects.value)
                if d_configs is not None:
                    oauth_configs = d_configs.get(OAuthConnect.K_OAUTH_CONN)
            else:
                oauth_configs = self.app_config.oauth_connect_configs
            k = AppConnect.search_default_key(oauth_configs)
            self.oauth_connect_key = k
            icps: Dict = self._init_connect_params.get(k)
            if icps is None:
                self.connect_params = self._configs\
                        [ConfigType.AppConnects.value]\
                        [OAuthConnect.K_OAUTH_CONN][k]
            else:
                self.connect_params = icps
                
            self.connect_type = self.connect_params.get(
                    ConfigKey.TYPE.value, '')
            self.connect_name = self.connect_params.get(
                    ConfigKey.NAME.value, '')
            self.connect_group = OAuthConnect.K_OAUTH_CONN
            self.connect_key = f"{OAuthConnect.K_OAUTH_CONN}.{k}"
            self.setup_connect(self.connect_params)



    @property
    def init_connect_params(self) -> Dict[str, object]:
        """
        Get a dictionary of initialized oauth connection config parameters.

        Returns:
            dict: initialized oauth connection config parameters.
        """   
        return OAuthConnect._init_connect_params
    

    def prepare_authorize_request(self, add_params: Dict = {}) -> Dict:
        pass


    def authorize_request(self, add_params: Dict = {}) -> Union[str, Dict]:
        """
        Construct an authorization request url. This method should be 
        overwritten by its child class.

        Args:
            add_params (Dict): additional parameters may be required for 
                authorization request.

        Returns:
            str: response json string.
        """
        return ""
    

    def process_authorize_response(self, json_ctx: Dict) -> Dict:
        """
        Construct an authorization request url. This method should be 
        overwritten by its child class.

        Args:
            json_ctx (Dict): json context in form of dictionary

        Returns:
            Dict: updated json context
        """
        return {}


    def prepare_grant_request(self, add_params: Dict = {}) -> Dict:
        pass


    def grant_request(self, add_params: Dict = {}) -> Dict:
        """
        Make grant request to get access token. This method should be
        overwritten by its child class.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """   
        return add_params
    

    def process_grant_response(self, json_ctx: Dict = {}) -> Dict:
        pass


    def generate_pkce_pair(self) -> Tuple[str, str]:
        """
        Generate pkce pair for Authorization code grant flow. For other
        types of flows, empty strings are returned

        Returns:
            tuple(str, str): code_verifier, code_challenge
        """
        pass



    @staticmethod
    def is_jwt(token: str) -> bool:
        try:
            jwt.get_unverified_header(token)
            return True
        except jwt.InvalidTokenError:
            return False
        

    def decode_token(
        self, 
        token_result: Dict, 
        token_types: List[str]
    ) -> Dict:
        """
        Decode access and refresh token.

        Example:

            ctx = oc.decode_token(
                tok_res, ["access_token", "refresh_token"]
            )

        Returns:
            Dict(str, obj): dictionary of token attributes
        """
        rt = {}
        for tok_t in token_types:
            self.logger.debug(
                f"OauthConnect.decode_token(): Token type [{tok_t}]"
            )
            token = token_result.get(tok_t)
            if not token:
                self.logger.error(
                    "OAuthConnect.decode_token(): No token is found "\
                    "in the provided response result!"
                )
            elif self.is_jwt(token):
                self.logger.debug(
                    f"OauthConnect.decode_token(): JWT token [{token}]."
                )
                dtok = jwt.decode(token, options={
                    "verify_signature": self.verify_signature}
                )
                rt[tok_t] = token
                rt[f"decoded_{tok_t}"] = dtok
            else:
                self.logger.warn(
                    f"OauthConnect.decode_token(): Not JWT token [{token}]!"
                )
                rt[tok_t] = token

        return rt


    def prepare_token_refresh(self) -> Dict:
        pass


    def refresh_token_request(self, add_params: Dict) -> Dict:
        """
        Make grant request to get refresh token. This method should be
        overwritten by its child class.

        Args:
            add_params (Dict): additional parameters may required for 
                grant token request.

        Returns:
            dict: a dictionary of grant token response results.
        """
        pass


    def setup_connect(self, params: Optional[Dict] = None):
        """
        Setup OAuthConnect with proper parameters initialized; it
        should be extended by its child class.

        Args:
            params (Dict): input parameters for connection setup.

        Returns:
            OAuthConnect: self with proper parameters initialized
        """
        self.connect_type = params.get(
                ConfigKey.TYPE.value, AppConfig.T_OAUTH_CODE)
        self.content_type = params.get(
                "content_type", 
                "application/x-www-form-urlencoded"
            )
        self.auth_request_url = params.get("auth_request_url", '')
        self.auth_response_fields = params.get("auth_response_fields", [])
        self.auth_response_type = params.get("auth_response_type")
        self.tenant_id = params.get("tenant_id", '')
        self.client_id = params.get("client_id", '')
        self.scope = params.get("scope", '')
        self.auth_response_errors = params.get(
                "auth_response_errors", 
                ["error", "error_description"]
            )
        self.grant_token_request_url = params.get(
                "grant_token_request_url", ''
            )
        self.grant_type = params.get("grant_type", "authorization_code")
        self.grant_token_response_fields = params.get(
                "grant_token_response_fields", ["access_token"]
            )
        self.client_secret_env = params.get(
                self.K_CLIENT_SECRET, "SNOWFLAKE_DEFAULT_APP_SECRET"
            )
        self.verify_signature = params.get("verify_signature", False)
        return self


    def init_connects(self) -> int:
        """
        Initialize a list of OAuthConnect configuration parameter configs.

        Returns:
            int: 0 - if it hasn't been initialized; other integer means 
                number of parameters dict have been initialized.
        """
        if not OAuthConnect._initialized:
            conn_group_dict = AppConfig.filter_group_key(
                OAuthConnect.K_OAUTH_CONN, 
                AppConnect.K_APP_CONN,
                self._configs
            )
            lst = conn_group_dict.get(OAuthConnect.K_INIT_LIST)
            if lst is not None and len(lst) > 0 :
                for oconn in lst:
                    dc: Dict = self._configs[AppConnect.K_APP_CONN]\
                        [OAuthConnect.K_OAUTH_CONN]
                    params: Dict = dc.get(oconn)
                    if params is None:
                        raise ValueError(
                            "OAuthConnect.init_connets(): Error - ["\
                            f"{oconn}] doesn't exist in the configuration!"
                        )
                    else:
                        self.init_connect_params[oconn] = params
                self.logger.debug(
                    f"OAuthConnect.init_connects(): Initialized OAuth - "\
                    f"{self.init_connect_params}"
                )
                OAuthConnect._initialized = True
        
        return len(self.init_connect_params)



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains Pipeline class for ML flow focusing on model
building, training, validation and testing in DEV domain.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import logging
from enum import Enum
from functools import wraps
from typing import Any, Callable, Dict, List, Optional, Tuple
from concurrent.futures import ThreadPoolExecutor, as_completed

from snowflake_ai.common import AppConfig, ConfigType, ConfigKey
from snowflake_ai.mlops import FlowContext


# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format='%(asctime)s [%(levelname)s]  %(message)s'
)



class TaskType(Enum):
    Setup = "setup"
    Ingest = "ingest"
    PreProcess = "preprocess"
    PreModel = "premodel"
    PreTrain = "pretrain"
    PrepData = "prepdata"
    PrepModel = "prepmodel"
    Learn = "learn"
    Train = "train"
    Validate = "validate"
    FineTune = "finetune"
    Test = "test"
    Predict = "predict"
    PostProcess = "postprocess"
    Evaluate = "evaluate"
    Deploy = "deploy"
    Serve = "serve"
    Feedback = "feedback"
    Monitor = "monitor"
    Retrain = "retrain"
    Cleanup = "cleanup"
    Custom = "custom"



class Step:

    _step_id = 0

    def __init__(
            self,
            func: Callable, 
            ctx: FlowContext,
            task_type: TaskType = None, 
            predecessor: 'Step' = None, 
            successor: 'Step' = None,
            iteration = 1
        ):
        self.func = func
        self.context = ctx
        self.task_type = task_type
        self.predecessor = predecessor
        self.successor = successor
        self.iteration = iteration
        self.error = False
        Step._step_id += 1
        self.step_id = Step._step_id



class Pipeline:
    """
    This class provides general setup for MLops flow in
    model building, training, validation and testing. 
    """

    def __init__(
            self, 
            pipeline_key: str,
            ctx: FlowContext = None,
            app_config: AppConfig = None
    ) -> None:
        self.logger = logging.getLogger(__name__)
        self.pipeline_key = pipeline_key
        _, self.pipeline_name = AppConfig.split_group_key(self.pipeline_key)
        if ctx is None:
            self.flow_context = FlowContext()
        else:
            self.flow_context = ctx
        self.flow_context.pipelines[pipeline_key] = {}
        self.pipeline_context = \
                self.flow_context.pipelines[pipeline_key]
        self.app_config = app_config
        self.steps : List[Step] = []
        self.step_tasks : List[TaskType] = []
        self.run: Callable = None
        self._init_pipeline_config()



    def _init_pipeline_config(self):
        conf_d = AppConfig.get_all_configs() if self.app_config is None \
                else self.app_config.get_all_configs()
        _, pp_config = AppConfig.get_group_item_config(
                self.pipeline_key,
                ConfigType.MLPipelines.value,
                conf_d
            )
        self.pipeline_type = pp_config.get(ConfigKey.TYPE.value, "")
        self.type = self.pipeline_type
        self.execution_mode = pp_config.get(ConfigKey.EXE_MODE.value, "")
        self.script = pp_config.get(ConfigKey.SCRIPT.value, "")
        self.class_name = pp_config.get(ConfigKey.CLASS_NAME.value, "")
        self.run_method = pp_config.get(ConfigKey.RUN.value, "")
        self.step_tasks = pp_config.get(ConfigKey.STEP_TASKS.value, [])

        rt = "" if self.app_config is None else self.app_config.root_path
        s = "" if self.app_config is None else self.app_config.script_home
        script_module = AppConfig.load_module(self.script, rt, s)
        if self.class_name:
            class_ = getattr(script_module, self.class_name)
            instance = class_()
            self.run = getattr(instance, self.run_method, "run") 
        else:
            self.run = getattr(script_module, self.run_method, "run")


    def execute(self):
        step_idx = 0
        task_idx = 0
        k_pp = self.pipeline_key
        ctx_pp: Dict[str, Dict] = self.flow_context.pipelines.get(k_pp, {})

        while task_idx < len(self.step_tasks):
        
            task_type = self.step_tasks[task_idx]
            step = self.steps[step_idx]
            step_name = step.func.__name__
            ctx_pp.setdefault(step_name, {FlowContext.T_STEP_ITER, 1})
            n_iter = ctx_pp[step_name][FlowContext.T_STEP_ITER]

            if step.task_type == task_type:
                self.logger.debug("Pipeline.execute(): "\
                        f"Executing Pipeline [{self.pipeline_key}];"\
                        f" Step [{step_name}]; Task index: {task_idx}")

                step.func(self.flow_context)

                if step.error:
                    self.logger.error("Pipeline.execute(): "\
                        f"Error in Step [{step_name}] in Pipeline "\
                        f"[{self.pipeline_key}]!")
                    return

                n_iter -= 1
                ctx_pp[step_name][FlowContext.T_STEP_ITER] = n_iter

                if n_iter > 0:
                    continue 
                elif step.successor:
                    step_idx = self.steps.index(
                        [s for s in self.steps if \
                                s.func.__name__ == step.successor][0]
                    )
                else:
                    step_idx += 1
                    task_idx += 1
            else:
                step_idx += 1


    def step(self, ctx: FlowContext, task_type, 
                predecessor=None, successor=None, iteration=1):
        def decorator(func):
            step_obj = Step(
                    func, self.flow_context, task_type, 
                    predecessor, successor)
            self.steps.append(step_obj)

            step_name = func.__name__
            ctx.pipelines.setdefault(self.pipeline_key, {})
            ctx.pipelines[self.pipeline_key][step_name] = \
                    {FlowContext.T_STEP_ITER: iteration}
            self.steps.append(Step(func, self.flow_context, task_type, 
                    predecessor, successor, iteration))

            return func
        return decorator


    def context(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func(
                self.flow_context,
                **kwargs
            )
        return wrapper


    @classmethod
    def flow(cls, func):
        def wrapper(
            obj: Pipeline
        ):
            _logger = logging.getLogger(__name__)
            ctx = FlowContext()
            if obj.flow_context is None:
                obj.flow_context = ctx
            else:
                ctx = obj.flow_context

            if ctx.direct_inputs is None:
                _logger.warn("Pipeline.flow(): Context input is None!")
                ctx.direct_inputs = {}
                return obj
            elif ctx.context_inputs is None:
                _logger.warn("Pipeline.flow(): Context input is None!")
                ctx.context_inputs = []
                return obj
            else:
                if ctx.direct_inputs and ctx.context_inputs:
                    ctx.context_inputs.append(ctx.direct_inputs)
                    func(obj)
                    ctx.context_inputs.append(ctx.outputs)
                    return obj
                elif ctx.context_inputs:
                    func(obj)
                    ctx.context_inputs.append(ctx.outputs)
                    return obj
                else:
                    ctx.context_inputs = []
                    func(obj)
                    ctx.context_inputs.append(ctx.outputs)
                    return obj
        return wrapper
    

    def run(self):
        ctx = self.flow_context
        if self.run is None:
            raise ValueError("Pipeline.run(): Misses main run function!")
        else:
            self.run(ctx, **ctx.direct_inputs)
        return 0



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains OAuthConnect class representing a generic OAuth
connection configuration.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


import sys
from typing import Optional, Dict, List
import logging
from snowflake.snowpark import Session

from snowflake_ai.common import  ConfigKey, ConfigType
from snowflake_ai.common import AppConfig, AppConnect
from snowflake_ai.common import DataConnect, OAuthConnect
from snowflake_ai.connect import SnowConnect, AuthCodeConnect
from snowflake_ai.connect import DeviceCodeConnect, ClientCredsConnect


class ConnectManager:
    """
    This class creates and manages application connects, including
    OAuthConnect, and DataConnect, etc.

    To use this class, create instance from AppConfig or extend
    it in a child class for creation of a specific OAuth connect
    instance, e.g., AuthCodeConnect as following :

        >>> from snowflake_ai.connect import AuthCodeConnect
        ...
        >>> ac = AppConfig("group_0.app_1")
        >>> conn = SecurityManager(ac).create_default_oauth_connect()
        >>> conn.grant_request()
    """

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    # initialization stats by app_key
    _init_d = {}

    # list of app connects by app key
    _app_conn_lst_d : Dict[str, List] = {} 
                                     

    @classmethod
    def create_default_oauth_connect(cls, app_config: AppConfig) \
            -> OAuthConnect:
        """
        Create default oauth connect.

        Returns:
            OAuthConnect: default oauth connect
        """   
        _initialized = cls._init_d.get(app_config.app_key)
        if _initialized is None or not _initialized:
            cls.create_app_connects(app_config)
        for c in cls._app_conn_lst_d[app_config.app_key]:
            if isinstance(c, OAuthConnect):
                return c
        return OAuthConnect(app_config.app_key, app_config)
    

    @classmethod
    def create_default_snow_connect(cls, app_config: AppConfig) \
            -> SnowConnect:
        """
        Create default snowflake connect.

        Returns:
            SnowConnect: default snowflake connect
        """   
        _initialized = cls._init_d.get(app_config.app_key)
        if _initialized is None or not _initialized:
            cls.create_app_connects(app_config)
        cls._logger.debug(
            f"ConnectManager.create_default_snow_connect(): App_key["\
            f"{app_config.app_key}]; List_of_connect[{cls._app_conn_lst_d}]."
        )
        for c in cls._app_conn_lst_d[app_config.app_key]:
            if isinstance(c, SnowConnect):
                return c
        return SnowConnect(app_config.app_key, app_config)
    

    @classmethod
    def create_app_connects(cls, app_config: AppConfig) -> List:
        """
        Create list of application connect instances

        Returns:
            List: application connect instances
        """ 
        if app_config is None:
            raise ValueError(
                "ConnectManager.create_app_connects(): AppConfig is required!"
            )
        
        _initialized = cls._init_d.get(app_config.app_key)
        if _initialized is None or not _initialized:
            cls._app_conn_lst_d[app_config.app_key] = []
            cls._logger.debug(
                "ConnectManager.create_app_connects(): "\
                f"App_connect_refs => {app_config.app_connect_refs}."
            )
            for ck in app_config.app_connect_refs:
                gk, k = AppConfig.split_group_key(ck)
                config = app_config.get_all_configs()\
                    [ConfigType.AppConnects.value][gk][k]
                
                if gk == OAuthConnect.K_OAUTH_CONN:
                    if config[ConfigKey.TYPE.value] == \
                            AppConfig.T_OAUTH_CODE:
                        ac = AppConnect.get_app_connects().get(ck)
                        if ac is None:
                            ac = AuthCodeConnect(ck, app_config)
                            AppConnect.get_app_connects()[ck] = ac
                        cls._app_conn_lst_d[app_config.app_key].append(ac)
                        cls._logger.debug(
                            "ConnectManager.create_app_connects(): "\
                            f"Initialize [{config[ConfigKey.TYPE.value]}] "\
                            f"OAuth Connection with Connect_Key[{ck}]."
                        )

                    elif config[ConfigKey.TYPE.value] == \
                            AppConfig.T_OAUTH_DEVICE:
                        ac = AppConnect.get_app_connects().get(ck)
                        if ac is None:
                            ac = DeviceCodeConnect(ck, app_config)
                            AppConnect.get_app_connects()[ck] = ac
                        cls._app_conn_lst_d[app_config.app_key].append(ac)
                        cls._logger.debug(
                            "ConnectManager.create_app_connects(): "\
                            f"Initialize [{config[ConfigKey.TYPE.value]}] "\
                            f"OAuth Connection with Connect_Key[{ck}]."
                        )

                    elif config[ConfigKey.TYPE.value] == \
                            AppConfig.T_OAUTH_CREDS:
                        ac = AppConnect.get_app_connects().get(ck)
                        if ac is None:
                            ac = ClientCredsConnect(ck, app_config)
                            AppConnect.get_app_connects()[ck] = ac
                        cls._app_conn_lst_d[app_config.app_key].append(ac)
                        cls._logger.debug(
                            "ConnectManager.create_app_connects(): "\
                            f"Initialize [{config[ConfigKey.TYPE.value]}] "\
                            f"OAuth Connection with Connect_Key[{ck}]."
                        )

                elif gk == DataConnect.K_DATA_CONN:
                    if config[ConfigKey.TYPE.value] == AppConfig.T_CONN_SNFLK:
                        ac = AppConnect.get_app_connects().get(ck)
                        if ac is None:
                            ac = SnowConnect(ck, app_config)
                            AppConnect.get_app_connects()[ck] = ac
                        cls._app_conn_lst_d[app_config.app_key].append(ac)
                        cls._logger.debug(
                            "ConnectManager.create_app_connects(): "\
                            f"Initialize [{config[ConfigKey.TYPE.value]}] "\
                            f"Data Connect with Connect_Key[{ck}]."
                        )
            
            cls._init_d[app_config.app_key] = True

        return cls._app_conn_lst_d[app_config.app_key]
    

    @classmethod
    def get_app_connects(cls, app_config: AppConfig) -> List:
        _initialized = cls._init_d.get(app_config.app_key)
        if _initialized is None or not _initialized:
            cls.create_app_connects(app_config)
        return cls._app_conn_lst_d[app_config.app_key]
    

    @classmethod
    def get_snowflake_service_session(cls, app_config: AppConfig) \
            -> Session:
        """
        Get snowflake session for the shared service data connection.
        Those connections should already be initialized based on the
        configuration.

        Returns:
            Session: snowflake session
        """   
        _initialized = cls._init_d.get(app_config.app_key)
        if _initialized is None or not _initialized:
            cls.create_app_connects(app_config)
        for c in cls._app_conn_lst_d[app_config.app_key]:
            if isinstance(c, SnowConnect):
                cls._logger.debug(
                    "ConnectManager.get_snowflake_service_session(): "\
                    f"Connect [{c.connect_key}]; Auth_type [{c.auth_type}]"
                )
                if (c.auth_type == AppConfig.T_AUTH_KEYPAIR) or \
                        (c.auth_type == AppConfig.T_AUTH_SNFLK) or\
                        (c.oauth_flow_type == AppConfig.T_OAUTH_CREDS):
                    return c.get_service_session()
        return None
    

    @classmethod
    def create_snowflake_user_session(
            cls, app_config: AppConfig, ctx: Dict
        ) -> Session:
        """
        Create user specific Snowflake session, typically through
        OAuth enabled SnowConnect.

        Returns:
            session: snowflake session
        """   
        _initialized = cls._init_d.get(app_config.app_key)
        if _initialized is None or not _initialized:
            cls.create_app_connects(app_config)
        for c in cls._app_conn_lst_d[app_config.app_key]:
            if isinstance(c, SnowConnect):
                if (c.auth_type == AppConfig.T_AUTH_OAUTH )or \
                        (c.auth_type == AppConfig.T_AUTH_EXT_BROWSER):                    
                    return c.create_user_session(ctx)
        return None


# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains AppConnect class representing a generic
application integration connection for data, security, etc.
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.4.0"


import sys
from typing import Optional, Dict, Union, Tuple
import logging

from snowflake_ai.common import ConfigType, ConfigKey, AppConfig



class AppConnect:
    """
    This class represents a generic application connection.
    Currently it is extended for oauth_connect, and
    data_connect.

    To use this class, create specific instances from AppConfig
    or extend it in a child class such as DataConnect. 
    """
    
    T_SNOWFLAKE_CONN = AppConfig.T_CONN_SNFLK
    T_AUTH_SNOWFLAKE = AppConfig.T_AUTH_SNFLK
    T_AUTH_KEYPAIR = AppConfig.T_AUTH_KEYPAIR
    T_AUTH_EXT_BROWSER = AppConfig.T_AUTH_EXT_BROWSER
    T_AUTH_OAUTH = AppConfig.T_AUTH_OAUTH
    K_APP_CONN = ConfigType.AppConnects.value

    SUPPORTED_CONNECT_GROUP_TYPES = [
        ConfigType.OAuthConnects.value, 
        ConfigType.DataConnects.value
    ]

    _logger = logging.getLogger(__name__)
    _logger.addHandler(logging.StreamHandler(sys.stdout))

    # all configurations
    _configs = {}

    # store Connect object reference
    _connects = {}


    def __init__(
            self, 
            connect_key : Optional[str] = None,
            app_config: AppConfig = None
        ):
        """
        Creat an AppConnect object.

        Args:
            connect_key: A string representing an AppConnect object;
                it can have the format of <app_connect_group>.<app_
                connect_name> ('<', '>' not included)
        """
        self.logger = AppConnect._logger
        self.connect_type, self.connect_name, self.auth_type = '', '', ''
        if app_config is not None:
            self.app_config = app_config
            self._configs = self.app_config.get_all_configs()
        else:
            self.app_config = None
            self._configs = AppConfig.get_all_configs()

        self.connect_group, self.connect_key = \
                AppConfig.split_group_key(connect_key)
        self.connect_key, self.connect_params = \
                AppConnect.load_connect_config(self.connect_key, self._configs)
        self.logger.debug(
                f"AppConnect.init(): Connect group [{self.connect_group}]; "\
                f"Connect key [{self.connect_key}]; "\
                f"App Connect Config Params => {self.connect_params}."
            )

        if self.connect_key and self.connect_params:
            self.app_connects[self.connect_key] = self

        if self.connect_params:
            self.auth_type = self.connect_params.get(
                    ConfigKey.AUTH_TYPE.value, '')
            self.type = self.connect_params.get(
                    ConfigKey.TYPE.value, '')
            self.connect_type = self.type
            self.oauth_flow_type = ""



    @staticmethod
    def load_connect_config(
        connect_key: str, 
        configs: Optional[Union[Dict, None]] = None
    ) -> Tuple[str, Dict]:
        """
        Load app connect configuration from overall configurations. 
        Connect key consists of <app_connect_group>.<app_
        connect_name> ('<', '>' not included), e.g.,
        "oauth_connects.auth_code_def".

        Args:
            connect_key (str): app connect key in form of 
                <app_connect_group>.<app_connect_name>
            configs (Dict): overall configuration dictionary

        Returns:
            Tuple[str, dict]: tuple of the connect key string matched in
            a form of group.connect_key and the dictionary of loaded
            app connect specific configurations.
        """
        if configs is not None and \
            configs.get(AppConnect.K_APP_CONN) is None:
            s = f"AppConnect.load_connect_config(): Error - "\
                "[app_connects] is missing!"
            AppConnect._logger.error(s)
            raise ValueError(s)
        if connect_key is None:
            return ('', {})
        
        if configs is None:
            configs = AppConfig.get_all_configs()

        rd = {}
        connect_key = connect_key.strip().lower()
        gk, ck = AppConfig.split_group_key(connect_key)
        if not gk and ck:
            k, rd = AppConfig.search_key_by_group(
                    ck, AppConnect.K_APP_CONN, configs
                )
        elif not gk and not ck:
            k, rd = '', {}
        elif gk and not ck:
            k, rd = f"{gk}.", AppConfig.filter_group_key(
                    gk, AppConnect.K_APP_CONN, configs
                )
        else:
            gs = dict(configs[AppConnect.K_APP_CONN]).get(gk)
            if gs is not None:
                k = f"{gk}.{ck}" 
                rd =  configs[AppConnect.K_APP_CONN][gk][ck] if dict(
                        configs[AppConnect.K_APP_CONN][gk]
                    ).get(ck) is not None else {}
            else:
                k, rd =  f"{gk}.{ck}", {}

        AppConfig._logger.debug(
            f"AppConnect.load_connect_config(): AppConnect[{k}] => {rd}"
        )
        return (k, rd)


    @staticmethod
    def search_default_key(data_dict: Dict[str, object]) -> str:
        """
        Search a dictionary for a key containing the substring "default". 
        If found, return that key. If not found, check if "_def", return
        the first found. If not found, check if "_0", retrun first found.
        If still not found any, sort the keys and return the first key 
        in the sorted list.

        Args:
            data_dict (dict): A dictionary with string keys.

        Returns:
            str, None: A key containing "default" or "_def". If found, or 
            the first key in the sorted list if not found.
        """
        if not data_dict: return None

        default_key = None
        sorted_keys = sorted(data_dict.keys())

        for k in sorted_keys:
            if (AppConfig.T_DEFAULT in k.lower()) or \
                    (AppConfig.T_DEF in k.lower()):
                default_key = k
                break

            if "_0" in k:
                default_key = k
                break

        return default_key or sorted_keys[0]
    

    @classmethod
    def get_app_connects(cls) -> Dict:
        """
        Get a dictionary of all currently loaded application connections.

        Returns:
            dict: a dictionary of all application connections.
        """  
        return AppConnect._connects


    @property
    def app_connects(self) -> Dict:
        """
        Get/set a dictionary of all currently loaded application connections.

        Returns:
            dict: a dictionary of all application connections.
        """   
        return AppConnect._connects
    

    def init_connects(self) -> int:
        """
        Initialize default AppConnect in each group. This generally should
        be overwritten by its child class
        """
        return len(AppConnect._connects)


    def is_oauth_saml_type(self) -> bool:
        """
        Return whether this application connect has auth_type == "oauth"
        or auth_type == "externalbrowser"
        """
        auth_t = self.auth_type
        return (auth_t == AppConfig.T_OAUTH) or \
                (auth_t == AppConnect.T_AUTH_EXT_BROWSER)



# Copyright (c) 2023, Tony Liu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# Use, reproduction and distribution of this software in source and 
# binary forms, with or without modification, are permitted provided that
# the License terms and conditions are met; you may not use this file
# except in compliance with the License. See the LICENSE file for details.

"""
This module contains FlowTest class for testing ML flow targeting
test/stage domain/environment
"""

__author__ = "Tony Liu"
__email__ = "tony.liu@yahoo.com"
__license__ = "Apache License 2.0"
__version__ = "0.3.0"


from snowflake_ai.mlops import Pipeline


class FlowTest(Pipeline):
    """
    This module contains FlowTest class for testing ML flow targeting
    test/stage domain/environment
    """

    def __init__(self) -> None:
        super().__init__()


    @Pipeline.flow
    def task_1(self):
        self.flow_context.direct_inputs = {"a": 1}
        return self
    

    @Pipeline.flow
    def task_2(self):
        self.flow_context.direct_inputs = {"a": 2}
        return self
    

    @Pipeline.flow
    def task_3(self):
        ctx = self.flow_context
        ns = [x["a"] for x in ctx.context_inputs if "a" in x]
        ctx.output = {"sum" : sum(ns)}
        return self
    

    def test_main(self):
        self.task_1().task_2().task_3()
        assert self.flow_context.outputs["sum"] == 3
    

if __name__ == '__main__':
    t = FlowTest()
    t.test_main()
